import numpy as np
import pandas as pd
from datetime import datetime
import matplotlib.pyplot as plt
plt.style.use('ggplot')
import seaborn as sns
from sklearn.cluster import KMeans
from sklearn.preprocessing import StandardScaler
from datetime import datetime
from sklearn.cluster import KMeans
from sklearn.preprocessing import StandardScaler
from imblearn.over_sampling import SMOTE
from sklearn.model_selection import train_test_split
from sklearn.metrics import f1_score
sns.set(color_codes=True)


def load_data(path="C://Users//C18244A//Documents//Datathon_Global_2018//data//20180920_Kaufland_case_IoT_and_predictive_maintenance.csv"):
    df = pd.read_csv(path, sep = ';', encoding='utf-8')

    # rename the columns
    df.columns = [x.split('.')[1] for x in df.columns.tolist()]

    schema = {
        "date_measurement": 'datetime64[ns]',
        "start_timestamp": 'datetime64[ns]',
        "end_timestamp": 'datetime64[ns]',
    }

    df = df.astype(schema)
    df['measurement_hour'] = df.start_timestamp.apply(lambda x: x.hour)
    df.drop(['start_timestamp', 'end_timestamp'], inplace=True, axis=1)
    return df


def save_each_machine_data(df):
    for machine in df.machine_name.unique().tolist():
        df[df.machine_name == machine].to_csv(f"..//data//{machine}.csv", index=False)


def dataprep(df):
    table = pd.pivot_table(df, values=['realvalue'], index=['machine_name', 'sensor_type', 
                                                            'date_measurement', 'measurement_hour'],
                           aggfunc=lambda x: np.percentile(x, 90))

    table.reset_index(inplace=True)

    #  Start the transpose process 1
    table = table.astype({'machine_name': str,
                          'sensor_type': str,
                          'date_measurement': str})

    table['new_index'] = table[['machine_name', 'sensor_type',
                                'date_measurement']].apply(lambda x: ')'.join(x), axis=1)

    table1 = table.pivot(index='new_index', columns='measurement_hour', values='realvalue')

    table2 = table1.reset_index()
    table2.columns = map(str, table2.columns)

    def filter_out(row):
        return [r for r in row if str(r) != 'nan']

    table2['list'] = list(map(filter_out, table2[[
                          x for x in table2.columns if x != 'new_index']].values.tolist()))

    table2.drop([x for x in table2.columns if x not in ['new_index', 'list']],
                inplace=True, axis=1)
    table2['number_of_observations'] = table2.apply(lambda x: len(x.list), axis=1)
    
    table3 = table2.new_index.apply(lambda x: pd.Series(x.split(')')))

    table3.columns = ['machine_name', 'sensor_type', 'date_measurement']
    table3['list'] = table2.list

    table3['number_of_observations'] = table2.number_of_observations

    max_number_of_observations_perDay = table3.number_of_observations.max()
    def good_list(l):
        if len(l) != max_number_of_observations_perDay:
            l += [None] * (max_number_of_observations_perDay - len(l))
        return l
    
    table3.list = table3.list.apply(good_list)
    for index in range(max_number_of_observations_perDay):
        table3['Num_measurement_' + str(index+1)] = table3.list.apply(lambda x: pd.Series(x[index]))
    table3.drop('list', axis=1, inplace=True)
    # table3.to_csv("..//data//pretty_data.csv", index=False)

    table3["date_measurement"] = pd.to_datetime(table3["date_measurement"])
    table3['day_of_observation'] = table3['date_measurement'].apply(
        lambda x: x.day)
    measure_columns = [i for i in table3.columns if i not in ['machine_name', 'sensor_type', 'date_measurement']]
    table3['max_measurement_of_day'] = table3[measure_columns].max(axis=1)
    table3.drop(measure_columns, axis=1, inplace=True)

    #  Start the transpose process 2
    table3['date_measurement'] = table3.date_measurement.astype(str)
    table3['new_index'] = table3[['machine_name', 'date_measurement']].apply(lambda x: ')'.join(x), axis=1)
    table4 = table3.pivot(index='new_index',
                          columns='sensor_type', 
                          values='max_measurement_of_day')
    table4.reset_index(inplace=True)
    table5 = table4.new_index.apply(lambda x: pd.Series([i for i in x.split(')')]))
    table5.columns = ['machine_name', 'date_measurement']

    needed_columns = ['drive_gear_V_eff', 'drive_gear_a_max',
                      'drive_motor_V_eff', 'drive_motor_a_max', 'drive_wheel_V_eff',
                      'drive_wheel_a_max', 'idle_wheel_V_eff', 'idle_wheel_a_max',
                      'lifting_gear_V_eff', 'lifting_gear_a_max', 'lifting_motor_V_eff',
                      'lifting_motor_a_max']
                    
    table5[needed_columns] = table4[needed_columns]
    return table5


def plot_heatmap(df):
    plt.figure(figsize=(15,15))
    sns.heatmap(df.corr(), annot=True)


def add_cluster_column(df):
    '''Generate cluster columns. The model is train by a machine.'''
    def find_clusters(data):
        X_train = data.iloc[:,2:]
        sc_X = StandardScaler()
        X_train_1 = sc_X.fit_transform(X_train)

        wcss = []
        for i in range(1, 20):
            kmeans = KMeans(n_clusters = i, init = 'k-means++', random_state = 42)
            kmeans.fit(X_train_1)
            wcss.append(kmeans.inertia_)
        kmeans = KMeans(n_clusters = 4, init = 'k-means++', random_state = 42)
        y_kmeans = kmeans.fit_predict(X_train_1)
        return y_kmeans.tolist()

    cluster_column = []
    for machine in df.machine_name.unique().tolist():
        cluster_column += find_clusters(df[df['machine_name'] == machine])

    df['cluster'] = pd.Series(cluster_column)
    return df


def add_time_gaps(df):
    '''Fill the time gaps in a given dataset.'''
    def fill_timestamp_gaps(df, machine_name, date_col='date_measurement'):
        full_period = pd.date_range(df[date_col].min(), 
                                    df[date_col].max()) 
        gaps = set(full_period).difference(set(df[date_col].tolist()))

        others = []
        for gap in gaps:
            others.append([machine_name, gap] + ([None] * (df.shape[1] - 2)))

        return df.append(pd.DataFrame(others, columns=df.columns), ignore_index=True)

    final_df = pd.DataFrame([], columns=df.columns)

    for machine in df.machine_name.unique().tolist():
        filtered_df = df[df.machine_name == machine]
        result = fill_timestamp_gaps(filtered_df, machine)
        final_df = final_df.append(result, ignore_index=True)

    return final_df


def import_final_error(final, path="Official_Failure 3.csv"):
    dateparse = lambda x: pd.datetime.strptime(x, '%d.%m.%Y')

    of = pd.read_csv(path, parse_dates = [0], date_parser = dateparse)

    final_err = final.merge(of, how = 'left', left_on = ['machine_name','date_measurement'], 
                            right_on = ['Machine', 'Date'])
                            
    final_err = final_err.drop(['drive_wheel_a_max_y', 'idle_wheel_V_eff_y',
    'idle_wheel_a_max_y', 'drive_gear_V_eff_y', 'drive_gear_a_max_y',
    'drive_motor_V_eff_y', 'drive_motor_a_max_y', 'lifting_gear_V_eff_y',
    'lifting_gear_a_max_y', 'lifting_motor_V_eff_y',
    'lifting_motor_a_max_y', 'Machine', 'Date'], axis = 1)

    final_err['Repair'] = final_err['drive_wheel_V_eff_y']
    final_err.drop('drive_wheel_V_eff_y', axis = 1, inplace = True)
    return final_err


def modeling(final, final_err):
    final['Missing_obs'] = np.where(final['drive_gear_a_max'] > 0, 0, 1)

    count = 0
    final['Number_of_days_missing'] = pd.Series([])

    for i in range(final.shape[0]):
        if final['Missing_obs'][i] == 1:
            count += 1
        else:
            for j in range(count):
                final['Number_of_days_missing'][i - j - 1] = count
            count = 0

    final_err_dedup = final_err.drop([570, 839, 903, 1196, 1440], axis=0)


    for sensor in ['drive_gear_V_eff',
                'drive_gear_a_max', 'drive_motor_V_eff', 'drive_motor_a_max',
                'drive_wheel_V_eff', 'drive_wheel_a_max', 'idle_wheel_V_eff',
                'idle_wheel_a_max', 'lifting_gear_V_eff', 'lifting_gear_a_max',
                'lifting_motor_V_eff', 'lifting_motor_a_max']:
        final['Rolling_mean_'+sensor] = final[sensor].rolling(7).mean()
        final['Rolling_std_'+sensor] = final[sensor].rolling(7).std()
        final['Rolling_max_'+sensor] = final[sensor].rolling(7).max()
        final['Rolling_min_'+sensor] = final[sensor].rolling(7).min()
        final['Rolling_diff_'+sensor] = final[sensor].diff()

    final_err_dedup[final_err_dedup['Repair'] == 1]
    final_err_dedup['Sum_rolling_mean_a'] = final_err_dedup['Rolling_mean_drive_gear_a_max'] + \
        final_err_dedup['Rolling_mean_drive_motor_a_max'] + \
        final_err_dedup['Rolling_mean_drive_wheel_a_max'] + \
        final_err_dedup['Rolling_mean_idle_wheel_a_max'] + \
        final_err_dedup['Rolling_mean_lifting_gear_a_max'] + \
        final_err_dedup['Rolling_mean_lifting_motor_a_max']

    final_err_dedup = final_err_dedup.set_index('date_measurement')

    final_err_dedup['Flag_high_rolling_mean'] = np.where(
        final_err_dedup['Sum_rolling_mean_a'] < 4*final_err_dedup['Sum_rolling_mean_a'].std(), 1, 0)

    final_err_dedup['Outcome'] = pd.Series([])

    for i in range(final_err_dedup.shape[0]):
        if final_err_dedup['Repair'][i] == 1:
            final_err_dedup['Outcome'][i] = 1
        elif final_err_dedup['Number_of_days_missing'][i] < 3:
            final_err_dedup['Outcome'][i] = 0
        elif (final_err_dedup['drive_gear_V_eff_x'][i] == final_err_dedup['drive_gear_V_eff_x'][i]) & \
            (final_err_dedup['drive_gear_V_eff_x'][i-1] != final_err_dedup['drive_gear_V_eff_x'][i-1]):
            final_err_dedup['Outcome'][i] = 0
        elif (final_err_dedup['Flag_high_rolling_mean'][i] == 0) \
                & (final_err_dedup['drive_gear_V_eff_x'][i] == final_err_dedup['drive_gear_V_eff_x'][i]):
            final_err_dedup['Outcome'][i] = 1
        else:
            final_err_dedup['Outcome'][i] = 0


    for machine in ['RBG1', 'RBG2', 'RBG3', 'RBG4', 'RBG5', 'RBG6', 'RBG7']:
        print(machine)
        # Predicting the Test set results
        vars()['y_'+machine] = final_err_dedup['Outcome'][final_err_dedup['machine_name']
                                                        == machine].fillna(0)
        vars()['X_'+machine] = final_err_dedup[final_err_dedup['machine_name'] == machine].drop(['Outcome',
                                                                                                'machine_name', 'Repair', 'Flag_high_rolling_mean'], axis=1).fillna(0)
        smt = SMOTE()
        vars()['X_enchanced_'+machine], vars()['y_enchanced_'+machine] = \
            smt.fit_sample(vars()['X_'+machine], vars()['y_'+machine])
        vars()['X_train_'+machine], vars()['X_test_'+machine],\
            vars()['y_train_'+machine], vars()['y_test_'+machine] = \
            train_test_split(vars()['X_enchanced_'+machine],
                            vars()['y_enchanced_'+machine], test_size=0.25, random_state=0)
        parameters = [{'C': [1, 10, 100, 1000], 'kernel': ['linear']},
                    {'C': [1, 10, 100, 1000], 'kernel': ['rbf'], 'gamma': [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9]}]
        grid_search = GridSearchCV(estimator=classifier,
                                   param_grid=parameters,
                                   scoring='accuracy',
                                   cv=10,
                                   n_jobs=-1)
        grid_search = grid_search.fit(
            vars()['X_train_'+machine], vars()['y_train_'+machine])
        best_parameters = grid_search.best_params_
        print(best_parameters)
        vars()['classifier_'+machine] = SVC(C=best_parameters['C'], kernel=best_parameters['kernel'],
                                            random_state=0)
    vars()['classifier_'+machine].fit(vars()['X_train_'+machine], vars()['y_train_'+machine])
